import json
from functools import wraps

import requests

from service_now_api_sdk.settings import (
    SERVICENOW_API_PASSWORD,
    SERVICENOW_API_TOKEN,
    SERVICENOW_API_USER,
    SERVICENOW_URL,
)


def headers_replace(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        if SERVICENOW_API_TOKEN:
            headers["Authorization"] = (f"Bearer {SERVICENOW_API_TOKEN}",)

        if kwargs.get("headers"):
            headers = {**headers, **kwargs.get["headers"]}

        kwargs["headers"] = headers

        return f(*args, **kwargs)

    return decorated_function


class Client:
    base_url = SERVICENOW_URL
    default_path = ""

    @headers_replace
    def __http_request(
        self,
        method: str,
        path: str,
        headers: dict = None,
        data=None,
        params: dict = None,
    ):
        if data is None:
            data = {}

        if params is None:
            params = {}

        if SERVICENOW_API_TOKEN:
            return requests.request(
                method=method,
                url=f"{self.base_url}/{path}",
                headers=headers,
                data=json.dumps(data),
                params=params,
            )

        if SERVICENOW_API_USER and SERVICENOW_API_PASSWORD:
            return requests.request(
                method=method,
                url=f"{self.base_url}/{path}",
                headers=headers,
                data=json.dumps(data),
                params=params,
                auth=(SERVICENOW_API_USER, SERVICENOW_API_PASSWORD),
            )

    def post(
        self, path: str, headers: dict = None, data: dict = None, params: dict = None
    ):
        return self.__http_request(
            method="POST", path=path, headers=headers, data=data, params=params
        )

    def get(self, path: str, headers: dict = None, params: dict = None):
        return self.__http_request(
            method="GET", path=path, headers=headers, params=params
        )

    def put(
        self, path: str, headers: dict = None, data: dict = None, params: dict = None
    ):
        return self.__http_request(
            method="PUT", path=path, headers=headers, data=data, params=params
        )

    def patch(
        self, path: str, headers: dict = None, data: dict = None, params: dict = None
    ):
        return self.__http_request(
            method="PATCH", path=path, headers=headers, data=data, params=params
        )

    def delete(self, path: str, headers: dict = None, data: dict = None):
        return self.__http_request(
            method="DELETE", path=path, headers=headers, data=data
        )



import os
from time import sleep

from service_now_api_sdk.sdk.servicenow.helpers.client import Client
from service_now_api_sdk.sdk.servicenow.helpers.query_builder import QueryBuilder
from service_now_api_sdk.sdk.servicenow.table.exceptions import (
    ManagerRetriveException,
    RecordFilterException,
    RecordRetriesException,
)


class BaseTableAPI:
    default_path = "api/now/table"
    table = ""
    http_client = Client()

    sysparm_display_value = False
    sysparm_exclude_reference_link = False
    sysparm_fields = None
    sysparm_query_no_domain = False
    sysparm_view = ""

    def __init__(self, table: str) -> None:
        super().__init__()
        self.table = table

    def view(self, view: str):
        """Render the response according to the specified UI view (overridden by sysparm_fields)

        Args:
            view (srtr): Specify UI view

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_view = view
        return self

    def exclude_reference_link(self, exclude: bool):
        """Exclude Table API links for reference fields (default: false)

        Args:
            exclude (bool): True to exclude Table API links for reference fields (default: false)

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_exclude_reference_link = exclude
        return self

    def display_value(self, display: str):
        """Return field display values (true), actual values (false), or both (all) (default: false).

        Args:
            display (str): Return field display values (default: false).

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_display_value = display
        return self

    def query_no_domain(self, no_domain: bool):
        """True to access data across domains if authorized (default: false)

        Args:
            no_domain (bool): Access data across domains if authorized (default: false)

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_query_no_domain = no_domain
        return self

    def only(self, fields: list):
        """List of fields to return in the response

        Args:
            fields (list): List of fields to return in the response

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_fields = ",".join(fields)
        return self

    def _get_params(self) -> dict:
        params = {}

        if self.sysparm_fields:
            params["sysparm_fields"] = self.sysparm_fields

        if self.sysparm_display_value:
            params["sysparm_display_value"] = self.sysparm_display_value

        if self.sysparm_exclude_reference_link:
            params[
                "sysparm_exclude_reference_link"
            ] = self.sysparm_exclude_reference_link

        if self.sysparm_view:
            params["sysparm_view"] = self.sysparm_view

        if self.sysparm_query_no_domain:
            params["sysparm_query_no_domain"] = self.sysparm_query_no_domain

        return params


class Records(BaseTableAPI):
    """Allows you to perform queries on existing tables
    Ref. link: https://docs.servicenow.com/bundle/quebec-application-development/page/integrate/inbound-rest/concept/c_TableAPI.html
    """

    query = None

    sysparm_limit: int = 500
    sysparm_offset: int = None
    sysparm_suppress_pagination_header: bool = False
    sysparm_query_category = None
    sysparm_no_count: bool = False
    __data = []

    def __init__(self, table: str):
        super().__init__(table=table)
        self.query = QueryBuilder()

    def suppress_pagination_header(self, supress: bool):
        """Supress pagination header (default: false)

        Args:
            supress (bool): True to supress pagination header (default: false)

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_suppress_pagination_header = supress
        return self

    def limit(self, limit: int):
        """The maximum number of results returned per page (default: 10000)

        Args:
            limit (int): The maximum number of results returned per page

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_limit = limit
        return self

    def offset(self, offset: int = 500):
        """The index of results returned per page (default: 500)

        Args:
            offset (int): The index of results returned per page

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_offset = offset
        return self

    def query_category(self, category: str):
        """Name of the query category (read replica category) to use for queries

        Args:
            view (srtr): Specify UI view

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_query_category = category
        return self

    def no_count(self, no_count: bool):
        """Do not execute a select count(*) on table (default: false)

        Args:
            no_domain (bool): True to execute a select count(*) on table (default: false)

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_no_count = no_count
        return self

    def _get_params(self) -> dict:
        params = super()._get_params()

        query = None
        if self.query._query:
            query = str(self.query)

        if query:
            params["sysparm_query"] = query

        if self.sysparm_limit != 500:
            params["sysparm_limit"] = self.sysparm_limit

        if self.sysparm_offset:
            params["sysparm_offset"] = self.sysparm_offset

        if self.sysparm_suppress_pagination_header:
            params[
                "sysparm_suppress_pagination_header"
            ] = self.sysparm_suppress_pagination_header

        if self.sysparm_query_category:
            params["sysparm_query_category"] = self.sysparm_query_category

        if self.sysparm_no_count:
            params["sysparm_no_count"] = self.sysparm_no_count

        return params

    def __request_helper(self, data=[], next_link="", retries=5):
        try:
            result = None
            params = self._get_params()
            current_data = []

            if next_link:
                result = self.http_client.get(next_link, params=params)
            else:
                result = self.http_client.get(
                    f"{self.default_path}/{self.table}", params=params
                )

            if result.status_code != 200:
                text = result.text
                raise RecordFilterException(text)

            if result.headers.get("content-type")[:16] == "application/json":
                current_data = result.json()
                data = data + current_data.get("result")

                if result.links.get("next"):
                    next_link = (
                        result.links.get("next", {})
                        .get("url", "")
                        .replace(f"{os.environ['SERVICENOW_URL']}/", "")
                    )
                    data = self.__request_helper(data, next_link=next_link)

            return data

        except Exception as e:
            if retries > 0:
                self.__request_helper(data, next_link=next_link, retries=retries - 1)
                print("Error: " + str(e))
                print("Retry in 30s")
                sleep(30)
            return

    def __request_helper_without_next_link(
        self, data: list[dict] = None, retries=5
    ) -> list[dict]:
        try:
            params = self._get_params()
            if not data:
                data = []

            if not self.sysparm_offset:
                self.sysparm_offset = 0

            result = self.http_client.get(
                path=f"{self.default_path}/{self.table}", params=params
            )

            if result.status_code != 200:
                text = result.text
                raise RecordFilterException(text)

            total_registers = int(result.headers["X-Total-Count"])

            current_data = result.json()
            data += current_data.get("result")

            if self.sysparm_offset + self.sysparm_limit < total_registers:
                self.sysparm_offset = self.sysparm_offset + self.sysparm_limit

                self.__request_helper_without_next_link(data=data, retries=retries)

            return data

        except Exception as e:
            if retries > 0:
                self.__request_helper(data=data, retries=retries - 1)
                print("Error: " + str(e))
                print("Retry in 30s")
                sleep(30)

            else:
                raise RecordRetriesException(e.__str__())

    def get(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        params = self._get_params()
        result = self.http_client.get(
            f"{self.default_path}/{self.table}", params=params
        )

        data = result.json()
        if result.status_code != 200:
            raise RecordFilterException(data)

        return data.get("result")

    def all(self):
        """[summary]

        Returns:
            [type]: [description]
        """

        if self.sysparm_suppress_pagination_header:
            return self.__request_helper_without_next_link()

        else:
            return self.__request_helper()


class Manager(BaseTableAPI):
    sysparm_input_display_value = None
    sysparm_suppress_auto_sys_field = None

    def input_display_value(self, input_display_value: bool):
        """Set field values using their display value (true) or actual value (false) (default: false)

        Args:
            input_display_value (bool): Set field values using their display value (true) or actual value (false) (default: false)

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_input_display_value = input_display_value
        return self

    def suppress_auto_sys_field(self, suppress_auto_sys_field: bool):
        """Suppress auto generation of system fields (default: false)

        Args:
            input_display_value (bool): True to suppress auto generation of system fields (default: false)

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_suppress_auto_sys_field = suppress_auto_sys_field
        return self

    def _get_params(self) -> dict:
        params = super()._get_params()

        if self.sysparm_suppress_auto_sys_field:
            params[
                "sysparm_suppress_auto_sys_field"
            ] = self.sysparm_suppress_auto_sys_field

        if self.sysparm_input_display_value:
            params["sysparm_input_display_value"] = self.sysparm_input_display_value

    def retrive(self, sys_id: str):
        result = self.http_client.get(
            f"{self.default_path}/{self.table}/{sys_id}", params=self._get_params()
        )

        data = result.json()
        if result.status_code != 200:
            raise ManagerRetriveException(data)
        return data

    def create(self, data: dict):
        result = self.http_client.post(
            f"{self.default_path}/{self.table}", data=data, params=self._get_params()
        )

        data = result.json()
        if result.status_code != 201:
            raise ManagerRetriveException(data)
        return data

    def delete(self, sys_id: str):
        result = self.http_client.delete(f"{self.default_path}/{self.table}/{sys_id}")

        data = result.json()
        if result.http_client.status_code != 200:
            raise ManagerRetriveException(data)
        return data

    def full_update(self, sys_id: str, data: dict):
        result = self.http_client.put(
            f"{self.default_path}/{self.table}/{sys_id}",
            data=data,
            params=self._get_params(),
        )

        data = result.json()
        if result.status_code != 200:
            raise ManagerRetriveException(data)
        return data

    def update(self, sys_id: str, data: dict):
        result = self.http_client.patch(
            f"{self.default_path}/{self.table}/{sys_id}",
            data=data,
            params=self._get_params(),
        )

        data = result.json()
        if result.status_code != 200:
            raise ManagerRetriveException(data)
        return data


class Vars(BaseTableAPI):
    __query = QueryBuilder()

    def __init__(self) -> None:
        super().__init__(table="sc_item_option_mtom")
        self.sysparm_limit = 500

    def _get_params(self) -> dict:
        params = super()._get_params()

        query = None
        if self.__query._query:
            query = str(self.__query)

        if query:
            params["sysparm_query"] = query

        return params

    def get_vars(self, by_field: str, data: str):
        self.__query.field(f"request_item.{by_field}").equals(data)
        self.__query.AND().field("request_item.sys_id").equals("sc_req_item.sys_id")
        self.__query.AND().field("sc_item_option_mtom.sc_item_option").equals(
            "sc_item_option.sys_id"
        )

        result = self.http_client.get(
            f"{self.default_path}/{self.table}", params=self._get_params()
        )

        data = result.json()
        if result.status_code != 200:
            raise RecordFilterException(data)

        return data


class ProducerServiceCatalog(Client):
    default_path = "api/sn_sc/servicecatalog/items"

    def store(self, catalog_id: str, variables: dict):
        path = f"{catalog_id}/submit_producer"
        result = self.post(f"{self.default_path}/{path}", data={"variables": variables})
        data = result.json()
        if result.status_code != 200:
            raise ManagerRetriveException(data)
        return data

    def store_task(self, catalog_id: str, variables: dict, sysparm_quantity: int = 1):
        """Open task ticket in servicenow
        Args:
            catalog_id (str, mandatory): service catalog sys id.
            variables (dict, mandatory): dictionary with variables values to open ticket.
            sysparm_quantity (int, optional): quantity of requisitions.

        Return:
            dict: data of ticket opened
        """
        path = f"{catalog_id}/order_now"
        payload = {"sysparm_quantity": sysparm_quantity, "variables": variables}
        result = self.post(f"{self.default_path}/{path}", data=payload)

        data = result.json()

        if result.status_code != 200:
            raise ManagerRetriveException(data)
        return data

    def update(self, sys_id: str, variables: dict):
        path = f"{sys_id}/submit_guide"
        result = self.put(f"{self.default_path}/{path}", data={"variables": variables})
        data = result.json()
        if result.status_code != 200:
            raise ManagerRetriveException(data)
        return data



import os
from time import sleep

from service_now_api_sdk.sdk.servicenow.attachments.exceptions import (
    DeleteAttachment,
    DownloadAttachment,
    RecordFilterException,
)
from service_now_api_sdk.sdk.servicenow.helpers.client import Client
from service_now_api_sdk.sdk.servicenow.helpers.query_builder import QueryBuilder


class BaseAttachmentsAPI:
    default_path = "api/now/attachment"
    http_client = Client()

    def __init__(self) -> None:
        super().__init__()


class Attachment(BaseAttachmentsAPI):
    """Allows you to download attachments files
    Ref. link: https://docs.servicenow.com/bundle/quebec-application-development/page/integrate/inbound-rest/concept/c_AttachmentAPI.html
    """

    sysparm_limit = None
    sysparm_offset = None
    query = None
    __file = []

    def __init__(self):
        super().__init__()
        self.query = QueryBuilder()

    def __request_helper(self, data=[], next_link="", retries=5):
        try:
            result = None
            params = self._get_params()
            current_data = []

            if next_link:
                result = self.http_client.get(next_link, params=params)
            else:
                result = self.http_client.get(f"{self.default_path}", params=params)

            if result.status_code != 200:
                text = result.text
                raise RecordFilterException(text)

            if result.headers.get("content-type")[:16] == "application/json":
                current_data = result.json()
                data = data + current_data.get("result")

                if result.links.get("next"):
                    next_link = (
                        result.links.get("next", {})
                        .get("url", "")
                        .replace(f"{os.environ['SERVICENOW_URL']}/", "")
                    )
                    data = self.__request_helper(data, next_link=next_link)

            return data
        except Exception as e:
            if retries > 0:
                self.__request_helper(data, next_link=next_link, retries=retries - 1)
                print("Error: " + str(e))
                print("Retry in 30s")
                sleep(30)
            return

    def _get_params(self) -> dict:
        params = {}

        query = None
        if self.query._query:
            query = str(self.query)

        if query:
            params["sysparm_query"] = query

        if self.sysparm_limit != 500:
            params["sysparm_limit"] = self.sysparm_limit

        if self.sysparm_offset:
            params["sysparm_limit"] = self.sysparm_limit

        return params

    def limit(self, limit: int):
        """The maximum number of results returned per page (default: 10000)

        Args:
            limit (int): The maximum number of results returned per page

        Returns:
            TableAPI: Return self class
        """
        self.sysparm_limit = limit
        return self

    def get_files_metadata(self):
        """Returns the metadata for multiple attachments.
        Ref. link: https://docs.servicenow.com/bundle/quebec-application-development/page/integrate/inbound-rest/concept/c_AttachmentAPI.html#attachment-GET

        Returns:
            dict: metadata registers result. Ref. link: https://docs.servicenow.com/bundle/quebec-application-development/page/integrate/inbound-rest/concept/c_AttachmentAPI.html#d1826795e895
        """
        return self.__request_helper()

    def retrieve_file_metadata(self, sys_id: str):
        """Retrieve a file metadata information
        Ref. link: https://docs.servicenow.com/bundle/quebec-application-development/page/integrate/inbound-rest/concept/c_AttachmentAPI.html#attachment-GET-file

        Args:
            sys_id (str, mandatory): Sys_id of the attachment record for which to retrieve the metadata.

        Returns:
            dict: metadata registers result. Ref. link: https://docs.servicenow.com/bundle/quebec-application-development/page/integrate/inbound-rest/concept/c_AttachmentAPI.html#attachment-GET-sys_id
        """
        result = self.http_client.get(path=f"{self.default_path}/{sys_id}")

        data = result.json()
        if result.status_code != 200:
            raise DownloadAttachment(data)

        return data.get("result")

    def download_file(self, sys_id: str, folder_path: str = "", file_name: str = None):
        """Download a servicenow file to locally storage.
        Ref. link: https://docs.servicenow.com/bundle/quebec-application-development/page/integrate/inbound-rest/concept/c_AttachmentAPI.html#attachment-GET-file

        Args:
            sys_id (str, mandatory): Sys_id of the attachment record from which to return binary data.
            folder_path (str, optional): Folder to store file.
            file_name (str, optional): File name to create locally file. If not defined, the raw name it will be used.

        Returns:
            file_path (str): path of file.
        """
        metadata = self.retrieve_file_metadata(sys_id=sys_id)

        if not file_name:
            file_name = metadata.get("file_name")

        result = self.http_client.get(path=f"{self.default_path}/{sys_id}/file")

        binary = result.content
        if result.status_code != 200:
            raise DownloadAttachment(result.json())

        file_path = f"{folder_path}/{file_name}"

        with open(file=file_path, mode="wb") as f:
            f.write(binary)

        return file_path

    def download_file_buffer(self, sys_id: str):
        """Returns the bytes array of file attachment with a specific sys_id value.
        Ref. link: https://docs.servicenow.com/bundle/quebec-application-development/page/integrate/inbound-rest/concept/c_AttachmentAPI.html#attachment-GET-file

        Args:
            sys_id (str, mandatory): Sys_id of the attachment record from which to return binary data.

        Returns:
            bytearray: the file buffer content.
        """
        result = self.http_client.get(path=f"{self.default_path}/{sys_id}/file")

        if result.status_code != 200:
            raise DownloadAttachment(result.json())

        binary = result.content

        return binary

    def delete_file(self, sys_id: str):
        """This method deletes the attachment with a specific sys_id value.
        Ref. link: https://docs.servicenow.com/bundle/quebec-application-development/page/integrate/inbound-rest/concept/c_AttachmentAPI.html#attachment-DELETE

        Args:
            sys_id (str, mandatory): Sys_id value of the attachment to delete.

        Returns:
            dict: metadata of file that has been deleted.
        """
        metadata = self.retrieve_file_metadata(sys_id=sys_id)

        result = self.http_client.delete(path=f"{self.default_path}/{sys_id}")

        if result.status_code != 204:
            raise DeleteAttachment(result.json)

        return metadata



import inspect
from datetime import timezone

import six

from service_now_api_sdk.sdk.servicenow.table.exceptions import (
    QueryEmpty,
    QueryExpressionError,
    QueryMissingField,
    QueryMultipleExpressions,
    QueryTypeError,
)


class QueryBuilder(object):
    """Query builder - for constructing advanced ServiceNow queries"""

    def __init__(self):
        self._query = []
        self.current_field = None
        self.c_oper = None
        self.l_oper = None

    def field(self, field):
        """Sets the field to operate on
        :param field: field (str) to operate on
        """

        self.current_field = field
        return self

    def order_descending(self):
        """Sets ordering of field descending"""

        self._query.append("ORDERBYDESC{0}".format(self.current_field))
        self.c_oper = inspect.currentframe().f_back.f_code.co_name
        return self

    def order_ascending(self):
        """Sets ordering of field ascending"""

        self._query.append("ORDERBY{0}".format(self.current_field))
        self.c_oper = inspect.currentframe().f_back.f_code.co_name
        return self

    def starts_with(self, starts_with):
        """Adds new `STARTSWITH` condition
        :param starts_with: Match field starting with the provided value
        """

        return self._add_condition("STARTSWITH", starts_with, types=[str])

    def ends_with(self, ends_with):
        """Adds new `ENDSWITH` condition
        :param ends_with: Match field ending with the provided value
        """

        return self._add_condition("ENDSWITH", ends_with, types=[str])

    def contains(self, contains):
        """Adds new `LIKE` condition
        :param contains: Match field containing the provided value
        """

        return self._add_condition("LIKE", contains, types=[str])

    def not_contains(self, not_contains):
        """Adds new `NOT LIKE` condition
        :param not_contains: Match field not containing the provided value
        """

        return self._add_condition("NOT LIKE", not_contains, types=[str])

    def is_empty(self):
        """Adds new `ISEMPTY` condition"""

        return self._add_condition("ISEMPTY", "", types=[str, int])

    def is_not_empty(self):
        """Adds new `ISNOTEMPTY` condition"""

        return self._add_condition("ISNOTEMPTY", "", types=[str, int])

    def equals(self, data):
        """Adds new `IN` or `=` condition depending on if a list or string was provided
        :param data: string or list of values
        :raise:
            - QueryTypeError: if `data` is of an unexpected type
        """

        if isinstance(data, six.string_types):
            return self._add_condition("=", data, types=[int, str])
        elif isinstance(data, list):
            return self._add_condition("IN", ",".join(map(str, data)), types=[str])

        raise QueryTypeError(
            "Expected value of type `str` or `list`, not %s" % type(data)
        )

    def not_equals(self, data):
        """Adds new `NOT IN` or `!=` condition depending on if a list or string was provided
        :param data: string or list of values
        :raise:
            - QueryTypeError: if `data` is of an unexpected type
        """

        if isinstance(data, six.string_types):
            return self._add_condition("!=", data, types=[int, str])
        elif isinstance(data, list):
            return self._add_condition("NOT IN", ",".join(data), types=[str])

        raise QueryTypeError(
            "Expected value of type `str` or `list`, not %s" % type(data)
        )

    def greater_than(self, greater_than):
        """Adds new `>` condition
        :param greater_than: str or datetime compatible object (naive UTC datetime or tz-aware datetime)
        :raise:
            - QueryTypeError: if `greater_than` is of an unexpected type
        """

        if hasattr(greater_than, "strftime"):
            greater_than = datetime_as_utc(greater_than).strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(greater_than, six.string_types):
            raise QueryTypeError(
                "Expected value of type `int` or instance of `datetime`, not %s"
                % type(greater_than)
            )

        return self._add_condition(">", greater_than, types=[int, str])

    def greater_than_or_equal(self, greater_than):
        """Adds new `>=` condition
        :param greater_than: str or datetime compatible object (naive UTC datetime or tz-aware datetime)
        :raise:
            - QueryTypeError: if `greater_than` is of an unexpected type
        """

        if hasattr(greater_than, "strftime"):
            greater_than = datetime_as_utc(greater_than).strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(greater_than, six.string_types):
            raise QueryTypeError(
                "Expected value of type `int` or instance of `datetime`, not %s"
                % type(greater_than)
            )

        return self._add_condition(">=", greater_than, types=[int, str])

    def less_than(self, less_than):
        """Adds new `<` condition
        :param less_than: str or datetime compatible object (naive UTC datetime or tz-aware datetime)
        :raise:
            - QueryTypeError: if `less_than` is of an unexpected type
        """

        if hasattr(less_than, "strftime"):
            less_than = datetime_as_utc(less_than).strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(less_than, six.string_types):
            raise QueryTypeError(
                "Expected value of type `int` or instance of `datetime`, not %s"
                % type(less_than)
            )

        return self._add_condition("<", less_than, types=[int, str])

    def less_than_or_equal(self, less_than):
        """Adds new `<=` condition
        :param less_than: str or datetime compatible object (naive UTC datetime or tz-aware datetime)
        :raise:
            - QueryTypeError: if `less_than` is of an unexpected type
        """

        if hasattr(less_than, "strftime"):
            less_than = datetime_as_utc(less_than).strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(less_than, six.string_types):
            raise QueryTypeError(
                "Expected value of type `int` or instance of `datetime`, not %s"
                % type(less_than)
            )

        return self._add_condition("<=", less_than, types=[int, str])

    def between(self, start, end):
        """Adds new `BETWEEN` condition
        :param start: int or datetime  compatible object (in SNOW user's timezone)
        :param end: int or datetime compatible object (in SNOW user's timezone)
        :raise:
            - QueryTypeError: if start or end arguments is of an invalid type
        """

        if hasattr(start, "strftime") and hasattr(end, "strftime"):
            dt_between = (
                'javascript:gs.dateGenerate("%(start)s")'
                "@"
                'javascript:gs.dateGenerate("%(end)s")'
            ) % {
                "start": start.strftime("%Y-%m-%d %H:%M:%S"),
                "end": end.strftime("%Y-%m-%d %H:%M:%S"),
            }
        elif isinstance(start, int) and isinstance(end, int):
            dt_between = "%d@%d" % (start, end)
        else:
            raise QueryTypeError(
                "Expected `start` and `end` of type `int` "
                "or instance of `datetime`, not %s and %s" % (type(start), type(end))
            )

        return self._add_condition("BETWEEN", dt_between, types=[str])

    def AND(self):
        """Adds an and-operator"""
        return self._add_logical_operator("^")

    def OR(self):
        """Adds an or-operator"""
        return self._add_logical_operator("^OR")

    def NQ(self):
        """Adds a NQ-operator (new query)"""
        return self._add_logical_operator("^NQ")

    def _add_condition(self, operator, operand, types):
        """Appends condition to self._query after performing validation
        :param operator: operator (str)
        :param operand: operand
        :param types: allowed types
        :raise:
            - QueryMissingField: if a field hasn't been set
            - QueryMultipleExpressions: if a condition already has been set
            - QueryTypeError: if the value is of an unexpected type
        """

        if not self.current_field:
            raise QueryMissingField("Conditions requires a field()")

        elif not type(operand) in types:
            caller = inspect.currentframe().f_back.f_code.co_name
            raise QueryTypeError(
                "Invalid type passed to %s() , expected: %s" % (caller, types)
            )

        elif self.c_oper:
            raise QueryMultipleExpressions("Expected logical operator after expression")

        self.c_oper = inspect.currentframe().f_back.f_code.co_name

        self._query.append(
            "%(current_field)s%(operator)s%(operand)s"
            % {
                "current_field": self.current_field,
                "operator": operator,
                "operand": operand,
            }
        )

        return self

    def _add_logical_operator(self, operator):
        """Adds a logical operator in query
        :param operator: logical operator (str)
        :raise:
            - QueryExpressionError: if a expression hasn't been set
        """

        if not self.c_oper:
            raise QueryExpressionError(
                "Logical operators must be preceded by an expression"
            )

        self.current_field = None
        self.c_oper = None

        self.l_oper = inspect.currentframe().f_back.f_code.co_name
        self._query.append(operator)
        return self

    def __str__(self):
        """String representation of the query object
        :raise:
            - QueryEmpty: if there's no conditions defined
            - QueryMissingField: if field() hasn't been set
            - QueryExpressionError: if a expression hasn't been set
        :return:
            - String-type query
        """

        if len(self._query) == 0:
            raise QueryEmpty("At least one condition is required")
        elif self.current_field is None:
            raise QueryMissingField("Logical operator expects a field()")
        elif self.c_oper is None:
            raise QueryExpressionError("field() expects an expression")

        return str().join(self._query)


def datetime_as_utc(date_obj):
    if date_obj.tzinfo is not None and date_obj.tzinfo.utcoffset(date_obj) is not None:
        return date_obj.astimezone(timezone.utc)
    return date_obj



from service_now_api_sdk.exceptions import ITSMException


class DownloadAttachment(ITSMException):
    pass


class RecordFilterException(ITSMException):
    pass


class DeleteAttachment(ITSMException):
    pass



from service_now_api_sdk.exceptions import ITSMException


class QueryTypeError(ITSMException):
    pass


class QueryMissingField(ITSMException):
    pass


class QueryEmpty(ITSMException):
    pass


class QueryExpressionError(ITSMException):
    pass


class QueryMultipleExpressions(ITSMException):
    pass


class RecordFilterException(ITSMException):
    pass


class RecordRetriesException(ITSMException):
    pass


class ManagerRetriveException(ITSMException):
    pass


class ManagerCreateException(ITSMException):
    pass


class ManagerDeleteException(ITSMException):
    pass


class ManagerFullUpdateException(ITSMException):
    pass


class ManagerUpdateException(ITSMException):
    pass


class ProducerSubmitException(ITSMException):
    pass



import os

SERVICENOW_URL = os.environ["SERVICENOW_URL"]
SERVICENOW_API_TOKEN = os.environ.get("SERVICENOW_API_TOKEN")

SERVICENOW_API_USER = os.environ.get("SERVICENOW_API_USER")
SERVICENOW_API_PASSWORD = os.environ.get("SERVICENOW_API_PASSWORD")
